window.IRacing = class IRacing
    constructor: (@requestParams=[], @requestParamsOnce=[], @fps=1, @server='127.0.0.1:8182', @readIbt=false) ->
        @data = {}
        @onConnect = null
        @onDisconnect = null
        @onUpdate = null

        @ws = null
        @reconnectTimeout = null

        @connected = false
        @firstTimeConnect = true

        @connect()

    connect: ->
        @ws = new WebSocket "ws://#{@server}/ws"
        @ws.onopen = => @onopen.apply @, arguments
        @ws.onmessage = => @onmessage.apply @, arguments
        @ws.onclose = => @onclose.apply @, arguments

    onopen: ->
        if @reconnectTimeout?
            clearTimeout @reconnectTimeout

        for k of @data
            delete @data[k]

        @ws.send JSON.stringify
            fps: @fps
            readIbt: @readIbt
            requestParams: @requestParams
            requestParamsOnce: @requestParamsOnce

    onmessage: (event) ->
        data = JSON.parse event.data

        # on disconnect
        if data.disconnected
            @connected = false
            if @onDisconnect
                @onDisconnect()

        # clear data on connect
        if data.connected
            for k of @data
                delete @data[k]

        # on connect or first time connect
        if data.connected or (@firstTimeConnect and not @connected)
            @firstTimeConnect = false
            @connected = true
            if @onConnect
                @onConnect()

        # update data
        if data.data
            keys = []
            for k, v of data.data
                keys.push(k)
                @data[k] = v
            if @onUpdate
                @onUpdate keys

    onclose: ->
        if @ws
            @ws.onopen = @ws.onmessage = @ws.onclose = null
        if @connected
            @connected = false
            if @onDisconnect
                @onDisconnect()
        @reconnectTimeout = setTimeout (=> @connect.apply @), 2000

    sendCommand: (command, args...) ->
        @ws.send JSON.stringify
            command: command
            args: args
