window.TwitchChat = class TwitchChat
    constructor: (@channel, @divId, @history=null, @host='127.0.0.1:8184') ->
        @conn = null
        @reconnectTimeout = null
        @emoticons = default: []
        @css = $('<style type="text/css">').appendTo 'head'
        @USER_COLORS = ['#ff0000', '#0000ff', '#008000', '#b22222', '#ff7f50', '#9acd32', '#ff4500', '#2e8b57', '#daa520', '#d2691e', '#5f9ea0', '#1e90ff', '#ff69b4', '#8a2be2', '#00ff7f']

    connect: ->
        @conn = new SockJS "http://#{@host}/chat"

        @conn.onopen = =>
            console.log 'connected'
            @conn.send JSON.stringify
                channel: @channel.toLowerCase()
                history: @history
            @checkVersionTimeout = setTimeout =>
                @onmessage()
            , 1000

        @conn.onmessage = (event) =>
            if @checkVersionTimeout?
                clearTimeout @checkVersionTimeout
                delete @checkVersionTimeout
            data = event.data
            if @onmessage
                @onmessage data.username, data.message, data.data

        @conn.onclose = =>
            console.log 'disconnected'
            @reconnectTimeout = setTimeout (=> @connect()), 2000

    loadBadges: (callback) ->
        req = $.getJSON "https://api.twitch.tv/kraken/chat/#{@channel}/badges?callback=?", (data) =>
            if data.error?
                console.log data.status, data.error, data.message
                setTimeout =>
                    @loadBadges callback
                , 1000
                return
            for type of data
                if type == '_links' or not data[type]?
                    continue
                @css.append "#{@divId} .#{type} { background-image: url(#{data[type].image}); }"
            if callback? then callback()

        req.fail =>
            console.log arguments
            setTimeout =>
                @loadBadges callback
            , 1000

    loadEmotes: (callback) ->
        req = $.getJSON "http://#{@host}/proxy/twitch_api/kraken/chat/emoticons", (data) =>
        # req = $.getJSON "http://#{@host}/proxy/twitch_api/kraken/chat/emoticon_images", (data) =>
        # req = $.getJSON "https://api.twitch.tv/kraken/chat/emoticon_images?callback=?", (data) =>
            if data.error?
                console.log data.status, data.error, data.message
                setTimeout =>
                    @loadEmotes callback
                , 1000
                return
            d = 0
            cssString = ''
            for a in data.emoticons
                if a.regex.match /^\w+$/
                    a.regex = new RegExp("\\b#{a.regex}\\b", 'g')
                else
                    a.regex = new RegExp(a.regex, 'g')
                for b in a.images
                    d += 1
                    b.html = "<span class='emo-#{d} emoticon'></span>"
                    imageObject =
                        image: b
                        regex: a.regex
                        css: @generateEmoticonCss b, d
                    if b.emoticon_set?
                        @emoticons[b.emoticon_set] ?= []
                        @emoticons[b.emoticon_set].push imageObject
                    else
                        @emoticons.default.push imageObject

            # for e in data.emoticons
            #     if e.code.match /^\w+$/
            #         e.code = new RegExp("\\b#{e.code}\\b", 'g')
            #     else
            #         e.code = new RegExp(e.code, 'g')
            #     cssString += @generateEmoticonCss e
            #     imageObject =
            #         regex: e.code
            #         html: "<span class='emo-#{e.id} emoticon'></span>"
            #     if e.emoticon_set?
            #         @emoticons[e.emoticon_set] ?= []
            #         @emoticons[e.emoticon_set].push imageObject
            #     else
            #         @emoticons.default.push imageObject

            if callback? then callback()

        req.fail =>
            console.log arguments
            setTimeout =>
                @loadEmotes callback
            , 1000

    clearChat: (username, strikeOut) ->
        if username?
            a = $("#{@divId} .chat-line[data-username=#{username}]")
        else
            a = $("#{@divId} .chat-line")
        if strikeOut
            a.css 'text-decoration': 'line-through'
        else
            a.remove()

    escape: (message) ->
        message.replace(/</g, '&lt;').replace(/>/g, '&gt;')

    emoticonize: (message, data, linkonize=false) ->
        emotes = []
        if data.emotes?
            for e in data.emotes.split '/'
                [id, places] = e.split ':'
                id = parseInt id
                for p in places.split ','
                    [start, end] = p.split '-'
                    emotes.push
                        id: id
                        start: parseInt start
                        end: parseInt end
            # emotes.sort (a, b) -> if a.start > b.start then -1 else if a.start < b.start then 1 else 0
            # prevStart = Infinity
            # for e in emotes
            #     message = message.substring(0, e.start) +
            #         "<img class=\"emoticon\" src=\"http://static-cdn.jtvnw.net/emoticons/v1/#{e.id}/1.0\">" +
            #         (if linkonize then @linkonize(@escape(message.substring(e.end + 1, prevStart))) else \
            #             @escape(message.substring(e.end + 1, prevStart))) +
            #         message.substring(prevStart)
            #     prevStart = e.start
            # message = (if linkonize then @linkonize(@escape(message.substring(0, prevStart))) else \
            #         @escape(message.substring(0, prevStart))) +
            #     message.substring(prevStart)
            emotes.sort (a, b) -> if a.start > b.start then 1 else if a.start < b.start then -1 else 0
            msg = []
            prevEmote = null
            for e in emotes
                msg.push message.substring((prevEmote?.end + 1) or 0, e.start), e
                prevEmote = e
            msg.push message.substring(prevEmote.end + 1)
            msg = msg.map (m, i) =>
                if i % 2
                    """<img class="emoticon" src="http://static-cdn.jtvnw.net/emoticons/v1/#{m.id}/1.0">"""
                else
                    if linkonize then @linkonize @escape m else @escape m
            message = msg.join ''
        else
            message = @escape message
            if linkonize
                message = @linkonize message
        message
        # matches = []
        # if data.emotes
        #     for set in data.emotes
        #         if not @emoticons[set]?
        #             continue
        #         for emote in @emoticons[set]
        #             if message.match emote.regex then matches.push emote
        # for emote in @emoticons.default
        #     if message.match emote.regex then matches.push emote
        # for emote in matches
        #     if emote.css?
        #         @css.append emote.css
        #         delete emote.css
        #     # message = message.replace emote.regex, emote.html
        #     message = message.replace emote.regex, emote.image.html
        # message

    generateEmoticonCss: (e, t) ->
        ".emo-#{t} {\
            background-image: url(#{e.url});\
            width: #{e.width}px;\
            height: #{e.height}px;\
            margin: #{12 - e.height >> 1}px 0;\
        }"
            # {if e.height > 16 then "margin: #{16 - e.height >> 1}px 0px;"}\
            # 28 -8
            # 26 -7
            # 18 -3

    # generateEmoticonCss: (e) ->
    #     ".emo-#{e.id} {\
    #         background-image: url(http://static-cdn.jtvnw.net/emoticons/v1/#{e.id}/1.0);\
    #     }"

    nameToColor: (name) ->
        acc = 0
        for i in [0...name.length]
            acc += name.charCodeAt(i)
        @USER_COLORS[acc % @USER_COLORS.length]

    calculateColorBackground: (color) ->
        # Converts HEX to YIQ to judge what color background the color would look best on
        color = String(color).replace /[^0-9a-f]/gi, ''
        if color.length < 6
            color = color[0] + color[0] + color[1] + color[1] + color[2] + color[2]

        r = parseInt(color.substr(0, 2), 16)
        g = parseInt(color.substr(2, 2), 16)
        b = parseInt(color.substr(4, 2), 16)
        yiq = ((r * 299) + (g * 587) + (b * 114)) / 1000
        yiq < 128

    calculateColorReplacement: (color, background) ->
        # Modified from http://www.sitepoint.com/javascript-generate-lighter-darker-color/
        inputColor = color
        rgb = '#'

        color = String(color).replace /[^0-9a-f]/gi, ''
        if color.length < 6
            color = color[0] + color[0] + color[1] + color[1] + color[2] + color[2]

        brightness = if background then '0.2' else '-0.5'

        for i in [0...3]
            c = parseInt(color.substr(i * 2, 2), 16)
            if c < 10 then c = 10
            c = Math.round(Math.min(Math.max(0, c + (c * brightness)), 255)).toString(16)
            rgb += ('00' + c).substr(c.length)

        if inputColor == rgb
            if background then "#ffffff" else "#000000"
        else
            rgb

    linkonize: (message) ->
        arr = message.split /((?:https?:\/\/)?(?:[-a-z0-9@:%_\+~#=]{1,256}\.)+[a-z]{2,6}(?:[\/?][-a-z0-9@:%_+.~#?&\/=]*)?)/ig
        if arr.length > 1
            for i in [1...arr.length] by 2
                link = arr[i]
                if link.search(/^https?/) == -1
                    link = 'http://' + link
                arr[i] = "<a href='#{link}' target='_blank'>#{arr[i]}</a>"
        arr.join ''

    filterBetbot: (username, message) ->
        if message.search(/^\s*!bet /i) == 0
            return true
        else if username in ['batmancave', @channel, 'the_turkey_bot']
            if message.search('TOP5: P1') == 0 then return true
            if message.search(/No one has more than [\d,]+/) == 0 then return true
            if message.search(/\s*Use "!bet win 100" to make a bet./) == 0 then return true
            if message.search(/, you have [\d,]+ .+/) != -1 then return true
            if message.search(/, you can bet at least half of your .+/) != -1 then return true
            if message.search(/, you can bet only all of your .+/) != -1 then return true
        false

    isAction: (message) ->
        if message.charCodeAt(0) == 1
            m = message.match(/\x01ACTION (.*)\x01/)
            if m.length > 1
                return [true, m[1]]
        [false, message]
